package lc.esp.sdk;

import com.anthonynsimon.url.URL;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import lc.mecha.util.StringAccumulatorV2;
import org.apache.activemq.command.ActiveMQQueue;
import org.apache.activemq.command.ActiveMQTopic;

import javax.jms.Destination;
import javax.jms.IllegalStateException;
import javax.jms.JMSException;
import javax.jms.Session;
import java.util.Locale;

/**
 * This class holds information about a destination for {@link ESPMessage}s.
 *
 * @author Alex Leigh
 * @since mk18 (GIPSY DANGER)
 */
public class ESPAddress {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(ESPAddress.class);
    private final String org;
    private final String domain;
    private final String service;
    private final String name;
    private final ESPAddressClass addressClass;
    private final ESPMessageClass messageClass;

    public ESPAddress(Destination dst) throws JMSException {
        logger.trace("Got destination: {} {}", dst.getClass().toString(), dst.toString());

        if (dst instanceof ActiveMQQueue) {
            ActiveMQQueue q = (ActiveMQQueue) dst;
            String[] path = q.getQueueName().split("\\.");
            this.org = path[0];
            this.domain = path[1];
            this.service = path[2];
            this.name = path[3];
            this.messageClass = ESPMessageClass.val(path[4]);
            this.addressClass = ESPAddressClass.QUEUE;
        } else if (dst instanceof ActiveMQTopic) {
            ActiveMQTopic q = (ActiveMQTopic) dst;
            String[] path = q.getTopicName().split("\\.");
            this.org = path[0];
            this.domain = path[1];
            this.service = path[2];
            this.name = path[3];
            this.messageClass = ESPMessageClass.val(path[4]);
            this.addressClass = ESPAddressClass.TOPIC;
        } else {
            throw new IllegalStateException("Unknown destination type.");
        }
    }

    public ESPAddress(String org, String domain, String service, String name,
                      ESPAddressClass serviceClass, ESPMessageClass destClass) {
        this.org = org;
        this.domain = domain;
        this.service = service;
        this.name = name;
        this.addressClass = serviceClass;
        this.messageClass = destClass;

    }

    /**
     * Parses a URL in string format.
     * <p>
     * queue://name
     * topic://name
     */
    public ESPAddress(String address) throws Exception {
        URL addrUrl = URL.parse(address);

        String[] host = addrUrl.getHostname().split("\\.");
        this.addressClass = ESPAddressClass.val(addrUrl.getScheme());
        this.org = host[0];
        this.domain = host[1];
        this.service = host[2];
        this.name = host[3];
        this.messageClass = ESPMessageClass.val(host[4]);
    }

    public String getDomain() {
        return domain;
    }

    public String getService() {
        return service;
    }

    public String getName() {
        return name;
    }

    public ESPAddressClass getAddressClass() {
        return addressClass;
    }

    public ESPMessageClass getMessageClass() {
        return messageClass;
    }

    public String getOrg() {
        return org;
    }

    @Override
    public String toString() {
        return "ESPDestination{" +
                "org='" + org + '\'' +
                ", domain='" + domain + '\'' +
                ", service='" + service + '\'' +
                ", name='" + name + '\'' +
                ", serviceClass=" + addressClass +
                ", destClass=" + messageClass +
                '}';
    }

    public String toOpenWireBase() {
        StringAccumulatorV2 sa = new StringAccumulatorV2(".");
        sa.push(org.toUpperCase(Locale.ROOT));
        sa.push(domain.toUpperCase(Locale.ROOT));
        sa.push(service.toUpperCase(Locale.ROOT));
        sa.push(name.toUpperCase(Locale.ROOT));
        return sa.asString();
    }

    /**
     * Returns this destination in OpenWire format. Note that the OpenWire address does not include the service
     * class, so the caller will need to deal with that appropriately.
     */
    public String toOpenWireAddress() {
        StringAccumulatorV2 sa = new StringAccumulatorV2(".");
        sa.push(org.toUpperCase(Locale.ROOT));
        sa.push(domain.toUpperCase(Locale.ROOT));
        sa.push(service.toUpperCase(Locale.ROOT));
        sa.push(name.toUpperCase(Locale.ROOT));
        sa.push(messageClass.toString().toUpperCase(Locale.ROOT));
        return sa.asString();
    }

    public String toMQTTAddress() {
        StringAccumulatorV2 sa = new StringAccumulatorV2("/");
        sa.push(org.toUpperCase(Locale.ROOT));
        sa.push(domain.toUpperCase(Locale.ROOT));
        sa.push(service.toUpperCase(Locale.ROOT));
        sa.push(name.toUpperCase(Locale.ROOT));
        sa.push(messageClass.toString().toUpperCase(Locale.ROOT));
        return sa.asString();
    }

    /**
     * Return this destination in STOMP format.
     */
    public String toStompAddress() {
        StringAccumulatorV2 sa = new StringAccumulatorV2("/");
        sa.push(addressClass.name().toLowerCase(Locale.ROOT));
        sa.push(org.toLowerCase(Locale.ROOT));
        sa.push(domain.toLowerCase(Locale.ROOT));
        sa.push(service.toLowerCase(Locale.ROOT));
        sa.push(name.toLowerCase(Locale.ROOT));
        sa.push(messageClass.toString().toUpperCase(Locale.ROOT));
        return sa.asString();
    }

    public Destination toDestination(Session session) throws JMSException {
        return switch (addressClass) {
            case QUEUE -> session.createQueue(toOpenWireAddress());
            case TOPIC -> session.createTopic(toOpenWireAddress());
            default -> throw new IllegalStateException("Unknown service class.");
        };
    }
}

