package lc.esp.sdk;

import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import lc.mecha.util.UniversalJob;
import lc.zero.sdk.ZeroActivationIndex;
import lc.zero.sdk.ZeroClient;
import lc.zero.sdk.ZeroServiceConfig;
import org.apache.activemq.ActiveMQConnectionFactory;

import javax.jms.Connection;
import javax.jms.IllegalStateException;
import javax.jms.JMSException;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;

/**
 * Enhanced Services Platform mk17 client.
 * <p>
 * This implementation is essentially a wrapper around the ActiveMQ JMS client. We are not adding much functionality
 * here other than some protocol semantics which are peculiar to ESP and ensuring that the right dependencies
 * are brought into the end-user's project.
 * <p>
 * In the future the client will also handle discovery and authentication. Note that authentication with ESP is
 * peculiar, so the regular username/password available via normal JMS will not work.
 *
 * @author Alex Leigh
 * @since mk17 (GIPSY DANGER)
 */
public class ESPClient implements AutoCloseable {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(ESPClient.class);
    private ActiveMQConnectionFactory connectionFactory;
    private Connection connection;
    private final ZeroClient zero = new ZeroClient();
    public static final String SVC_ESP = "esp";

    public ESPClient() throws KeyStoreException, NoSuchAlgorithmException, KeyManagementException {
        UniversalJob.banner(logger, "Enhanced Services Platform/SDK mk18 (GIPSY DANGER)");
    }

    /**
     * Returns the ESPClient's {@link ZeroClient} to the caller. The caller can use this to determine other
     * application-specific service configuration. If <code>start()</code> has not been called, this method
     * will return <code>null</code>.
     */
    public ZeroClient getZero() {
        return zero;
    }

    /**
     * Start the ESP client. This will perform Zero activation. The method blocks until the zero activation
     * is successful.
     */
    public ESPClient start() throws Exception {
        new Thread(zero).start();
        ZeroActivationIndex zai;

        zai = zero.getZai();

        ZeroServiceConfig espCfg = zai.readConfig(SVC_ESP);

        if (espCfg == null) {
            throw new IllegalStateException("No service configuration found in ZERO.");
        }

        JSONObject connections = espCfg.getCfg().getJSONObject("connections");
        String openWireUrl = connections.getString("openwire");

        logger.info("Learned configuration from ZERO. [openwire: {}]", openWireUrl);

        connectionFactory = new ActiveMQConnectionFactory(openWireUrl);
        connectionFactory.setUseAsyncSend(true);
        connection = getConnectionFactory().createConnection();
        connection.start();
        return this;
    }

    /**
     * Return the connection factory associated with this client.
     */
    public ActiveMQConnectionFactory getConnectionFactory() {
        return connectionFactory;
    }

    @Override
    public void close() throws Exception {
        if (connection != null) connection.close();
    }

    public ESPSession createSession() throws JMSException {
        return new ESPSession(connection);
    }
}