package lc.esp.sdk;

import lc.mecha.json.JSONArray;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.apache.activemq.command.ActiveMQBytesMessage;

import javax.jms.IllegalStateException;
import javax.jms.*;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Objects;

public class ESPMessage {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(ESPMessage.class);
    public static final String KEY_PAYLOAD = "payload";
    public static final String KEY_PARAMETERS = "parameters";
    public static final String KEY_NAME = "name";
    public static final String KEY_VALUE = "value";

    private JSONObject payload;
    private ESPAddress replyTo;
    private ESPAddress destination;
    private String correlationId;
    private final Map<String, Serializable> parameters = new HashMap<>();

    public ESPMessage(Message msg) throws JMSException {
        logger.debug("Building from message: {}", msg);

        Destination dest = msg.getJMSDestination();
        if (dest != null) {
            destination = new ESPAddress(dest);
        }

        correlationId = msg.getJMSCorrelationID();

        Destination dst = msg.getJMSReplyTo();
        if (dst != null) {
            replyTo = new ESPAddress(dst);
        }

        for (Iterator<String> it = msg.getPropertyNames().asIterator(); it.hasNext(); ) {
            String param = it.next();
            parameters.put(param, (Serializable) msg.getObjectProperty(param));
        }
        if (msg instanceof TextMessage) {
            payload = new JSONObject(((TextMessage) msg).getText());
        } else if (msg instanceof ActiveMQBytesMessage) {
            String s = new String(((ActiveMQBytesMessage) msg).getContent().getData());
            payload = new JSONObject(s);
        } else {
            throw new IllegalStateException("Unknown message type: " + msg.getClass());
        }
    }

    /**
     * Serialize the entire message into JSON.
     */
    public JSONObject toJson() {
        JSONObject o = new JSONObject();
        o.put(KEY_PAYLOAD, payload);
        JSONArray arr = new JSONArray();
        for (Map.Entry<String, Serializable> param : parameters.entrySet()) {
            arr.put(new JSONObject().put(KEY_NAME, param.getKey()).put(KEY_VALUE, param.getValue()));
        }
        if (arr.length() > 0) o.put(KEY_PARAMETERS, arr);
        return o;
    }

    public JSONObject getPayload() {
        return payload;
    }

    /**
     * Set a parameter in the message. This parameter may later be used to filter
     * messages during the subscription process (by creating a selector).
     * <p>
     * All parameters which begin with underscore are reserved by ESP.
     */
    public void setParameter(String name, Serializable value) {
        parameters.put(name, value);
    }

    public ESPMessage() {

    }

    public Serializable getParameter(String name) {
        return parameters.get(name);
    }

    public ESPMessage(JSONObject payload, ESPAddress replyTo, String correlationId) {
        this.payload = payload;
        this.replyTo = replyTo;
        this.correlationId = correlationId;
    }

    public void setPayload(JSONObject payload) {
        this.payload = payload;
    }

    public ESPMessage(JSONObject payload) {
        this(payload, null, null);
    }

    public String getCorrelationId() {
        return correlationId;
    }

    public Map<String, Serializable> getParameters() {
        return parameters;
    }

    @Override
    public String toString() {
        return "ESPMessage{" +
                "payload=" + payload +
                ", replyTo=" + replyTo +
                ", destination=" + destination +
                ", correlationId='" + correlationId + '\'' +
                ", parameters=" + parameters +
                '}';
    }

    public ESPAddress getReplyTo() {
        return replyTo;
    }

    public ESPAddress getDestination() {
        return destination;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        ESPMessage that = (ESPMessage) o;
        return Objects.equals(payload, that.payload);
    }

    @Override
    public int hashCode() {
        return Objects.hash(payload);
    }

    public void setReplyTo(ESPAddress replyTo) {
        this.replyTo = replyTo;
    }

    public void setCorrelationId(String correlationId) {
        this.correlationId = correlationId;
    }
}
