package lc.esp.sdk.telemetry;

import lc.esp.sdk.ESPMessage;
import lc.mecha.json.JSONArray;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;

import java.io.Serializable;
import java.time.Instant;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * A telemetry frame contains zero or more symbols which are grouped together into a single "reading".
 * <p>
 * If symbols happen to be inter-related, for example, a X-Y-Z positional vector, then they are said to be of the
 * same reading if they are within the same frame. The frame constitutes a particular instant of time.
 *
 * @author Alex Leigh
 * @since mk18 (GIPSY DANGER)
 */
public class TelemetryFrame {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(TelemetryFrame.class);
    public static final String KEY_TIME = "time";
    public static final String KEY_TIME_NS = "time_ns";
    public static final String KEY_SYMBOLS = "symbols";
    public static final String PREFIX_TLM = "esp.tlm.";
    private final HashSet<TelemetrySymbol> symbols = new HashSet<>();
    private final HashMap<String, Serializable> tags = new HashMap<>();
    private final Instant time;

    /**
     * Create a TelemetryFrame object from a {@link ESPMessage}.
     * <p>
     * All ESP message parameters that begin with "esp.tlm." will be copied into the TelemetryFrame with the TLM.
     * removed.
     *
     * @param msg The message.
     */
    public TelemetryFrame(ESPMessage msg) {
        JSONObject json = msg.getPayload();

        logger.trace("Payload: {}", json);

        if (json.has(KEY_TIME)) {
            time = Instant.ofEpochMilli(json.optLong(KEY_TIME));
        } else {
            // This is actually poor because we don't know how long
            // the thing has been sitting in the queue
            time = Instant.now();
        }

        JSONArray arr = json.getJSONArray(KEY_SYMBOLS);
        for (int i = 0; i < arr.length(); i++) {
            JSONObject symJson = arr.getJSONObject(i);
            symbols.add(new TelemetrySymbol(symJson));
        }

        for (Map.Entry<String, Serializable> param : msg.getParameters().entrySet()) {
            if (param.getKey().startsWith(PREFIX_TLM)) {
                String tagName = param.getKey().substring(PREFIX_TLM.length());
                tags.put(tagName, param.getValue());
            }
        }
    }

    public TelemetryFrame(Instant time) {
        this.time = time;
    }

    public Instant getTime() {
        return time;
    }

    public JSONObject toJson() {
        JSONObject frameJson = new JSONObject();
        frameJson.put(KEY_TIME, time.toEpochMilli());
        frameJson.put(KEY_TIME_NS, time.getNano());
        JSONArray symbolsArray = new JSONArray();
        for (TelemetrySymbol s : symbols) {
            symbolsArray.put(s.toJson());
        }
        if (symbolsArray.length() > 0) {
            frameJson.put(KEY_SYMBOLS, symbolsArray);
        }
        return frameJson;
    }

    /**
     * Convert this TelemetryFrame to an {@link ESPMessage} in the standard format. This will copy all of the
     * telemetry frame tags to the message parameters while prepending the keys with "esp.tlm.".
     */
    public ESPMessage toMessage() {
        ESPMessage msg = new ESPMessage();
        for (Map.Entry<String, Serializable> parameter : tags.entrySet()) {
            msg.getParameters().put(PREFIX_TLM + parameter.getKey(), parameter.getValue());
        }
        msg.setPayload(toJson());
        return msg;
    }

    public HashMap<String, Serializable> getTags() {
        return tags;
    }

    @Override
    public String toString() {
        return "TelemetryFrame{" +
                "symbols=" + symbols +
                ", tags=" + tags +
                ", time=" + time +
                '}';
    }

    /**
     * Return the symbols associated with this frame.
     */
    public Set<TelemetrySymbol> getSymbols() {
        return symbols;
    }

    public void addSymbol(TelemetrySymbol symbol) {
        // This works on the principal that equals() is implemented properly
        symbols.add(symbol);
    }

}
