package lc.esp.sdk.telemetry;

import lc.mecha.json.JSONObject;

import java.io.Serializable;
import java.util.Objects;

/**
 * This class holds data about a particular telemetry symbol reading.
 * <p>
 * Symbols are meant to be aggregated into {@link TelemetryFrame}s. The frame contains important meta data
 * about the symbol, including the time the symbol was valid for.
 * <p>
 * When designing symbols, keep in mind that the value of the symbol should never be a set, array, or object. The
 * value should be primitive. In cases where a reading is a vector, each part should be split up into an independent
 * symbol, and then all those symbols can be represented in a frame to indicated they occurred at the
 * same moment/reading.
 *
 * @author Alex Leigh
 * @since mk18 (GIPSY DANGER)
 */
public class TelemetrySymbol {
    public static final String KEY_NAME = "name";
    public static final String KEY_VALUE = "value";
    public static final String KEY_UOM = "uom";
    public static final String KEY_LOS = "los";
    public static final String KEY_PRIMARY = "primary";
    private final String name;
    private final Serializable value;
    private final String uom;
    private final boolean isLos;
    private final boolean isPrimary;

    public TelemetrySymbol(JSONObject obj) {
        name = obj.getString(KEY_NAME);
        value = (Serializable) obj.get(KEY_VALUE);
        uom = obj.optString(KEY_UOM, null);
        isLos = obj.getBoolean(KEY_LOS);
        isPrimary = obj.getBoolean(KEY_PRIMARY);
    }

    public TelemetrySymbol(String name, Serializable value, boolean isPrimary) {
        this.name = name;
        this.value = value;
        this.uom = null;
        this.isLos = false;
        this.isPrimary = isPrimary;
    }

    public TelemetrySymbol(String name, Serializable value, String uom, boolean isLos, boolean isPrimary) {
        this.name = name;
        this.value = value;
        this.uom = uom;
        this.isLos = isLos;
        this.isPrimary = isPrimary;
    }

    public JSONObject toJson() {
        JSONObject obj = new JSONObject();
        obj.put(KEY_NAME, name);
        obj.put(KEY_VALUE, value);
        obj.put(KEY_UOM, uom);
        obj.put(KEY_LOS, isLos);
        obj.put(KEY_PRIMARY, isPrimary);
        return obj;
    }

    @Override
    public String toString() {
        return "TelemetrySymbol{" +
                "name='" + name + '\'' +
                ", value=" + value +
                ", uom='" + uom + '\'' +
                ", isLos=" + isLos +
                '}';
    }

    /**
     * Returns <code>true</code>> of the symbol is in the Loss Of Signal (LOM) state.
     *
     *  Symbols can be independently in the LOS state; this indicates that the source for the symbol has become unlocked.
     *  The symbol may or may not contain a value; if it does, that value indicates the last known reading on the symbol
     *  before the LOS event occurred.
     */
    public boolean isLos() {
        return isLos;
    }

    /**
     * Return the unit-of-measure, if any, for this symbol.
     */
    public String getUom() {
        return uom;
    }

    /**
     * Return the name for this symbol.
     */
    public String getName() {
        return name;
    }

    /**
     * Return the value of this symbol.
     */
    public Serializable getValue() {
        return value;
    }

    /**
     * Returns <code>true</code> if this symbol value is primary. A value is considered to be primary if it is a direct
     * raw value such as that read directly from a sensor. A value is considered to be secondary if it is the product
     * of some calculation, including aggregations, sampling, or other formulas.
     */
    public boolean isPrimary() {
        return isPrimary;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        TelemetrySymbol that = (TelemetrySymbol) o;
        return Objects.equals(name, that.name);
    }

    @Override
    public int hashCode() {
        return Objects.hash(name);
    }
}
