package lc.minecraft.link;

import lc.esp.sdk.*;
import lc.esp.sdk.telemetry.TelemetryFrame;
import lc.esp.sdk.telemetry.TelemetrySymbol;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.bukkit.Location;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.player.PlayerMoveEvent;

import java.time.Instant;

/**
 * This class implements a bukkit {@link Listener} which is primarily responsible for receiving events from the server
 * and broadcasting a telemetry signal in realtime to ESP.
 *
 * @author Alex Leigh
 * @since mk2
 */
public class BukkitEventHandler implements Listener {
    public static final String KEY_LOC_TO_X = "lc.mc.loc.dst.x";
    public static final String KEY_LOC_TO_Y = "lc.mc.loc.dst.y";
    public static final String KEY_LOC_TO_Z = "lc.mc.loc.dst.z";
    public static final String KEY_LOC_TO_YAW = "lc.mc.loc.dst.yaw";
    public static final String KEY_LOC_TO_PITCH = "lc.mc.loc.dst.pitch";
    public static final String KEY_LOC_FROM_X = "lc.mc.loc.src.x";
    public static final String KEY_LOC_FROM_Y = "lc.mc.loc.src.y";
    public static final String KEY_LOC_FROM_Z = "lc.mc.loc.src.z";
    public static final String KEY_LOC_FROM_YAW = "lc.mc.loc.src.yaw";
    public static final String KEY_LOC_FROM_PITCH = "lc.mc.loc.src.pitch";
    public static final String KEY_EVENT_TYPE = "lc.mc.evt.typ";
    public static final String KEY_BLOCK_TYPE = "lc.mc.blk.typ";
    public static final String KEY_SERVER = "lc.mc.server";
    public static final String KEY_PLAYER_NAME = "lc.mc.player_name";
    public static final String KEY_WORLD = "lc.mc.world";
    public static final String EVENT_BLOCK_BREAK = "lc.mc.evt.blk_break";
    public static final String EVENT_BLOCK_PLACE = "lc.mc.evt.blk_place";
    public static final String EVENT_PLAYER_MOVE = "lc.mc.evt.blk_brk.player_mv";
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(BukkitEventHandler.class);
    private final ESPClient esp;
    private final ESPProducer producer;
    private final AsyncSender sender;
    private final String serverName;

    public BukkitEventHandler(String serverName) throws Exception {
        this.serverName = serverName;
        esp = new ESPClient();
        esp.start();
        ESPAddress telemDest = new ESPAddress("leighco", "prod",
                "minecraft", "neweden", ESPAddressClass.TOPIC,
                ESPMessageClass.TELEMETRY);
        ESPSession session = esp.createSession();
        producer = session.createProducer(telemDest);
        sender = new AsyncSender(producer);
        new Thread(sender).start();
    }

    private void addDstLocation(TelemetryFrame values, Location loc) {
        addValue(values, KEY_LOC_TO_X, Double.toString(loc.getX()));
        addValue(values, KEY_LOC_TO_Y, Double.toString(loc.getY()));
        addValue(values, KEY_LOC_TO_Z, Double.toString(loc.getZ()));
    }

    private void addSrcLocation(TelemetryFrame values, Location loc) {
        addValue(values, KEY_LOC_FROM_X, Double.toString(loc.getX()));
        addValue(values, KEY_LOC_FROM_Y, Double.toString(loc.getY()));
        addValue(values, KEY_LOC_FROM_Z, Double.toString(loc.getZ()));
    }

    private void addDstAngle(TelemetryFrame values, Location loc) {
        // FIXME: Should set UOM to degrees!
        addValue(values, KEY_LOC_TO_YAW, Double.toString(loc.getYaw()));
        addValue(values, KEY_LOC_TO_PITCH, Double.toString(loc.getPitch()));
    }

    private void addSrcAngle(TelemetryFrame values, Location loc) {
        // FIXME: Should set UOM to degrees!
        addValue(values, KEY_LOC_FROM_YAW, Double.toString(loc.getYaw()));
        addValue(values, KEY_LOC_FROM_PITCH, Double.toString(loc.getPitch()));
    }

    private void addValue(TelemetryFrame values, String key, String value) {
        values.addSymbol(new TelemetrySymbol(key, value, true));
    }

    @EventHandler(ignoreCancelled = true)
    public void onBlockBreak(BlockBreakEvent event) {
        try {
            TelemetryFrame frame = new TelemetryFrame(Instant.now());
            addDstLocation(frame, event.getBlock().getLocation());
            addSrcLocation(frame, event.getPlayer().getLocation());
            addDstLocation(frame, event.getBlock().getLocation());
            addSrcLocation(frame, event.getPlayer().getLocation());

            ESPMessage msg = new ESPMessage();
            msg.setParameter(KEY_SERVER, serverName);
            msg.setParameter(KEY_PLAYER_NAME, event.getPlayer().getName());
            msg.setParameter(KEY_WORLD, event.getBlock().getWorld().getName());
            msg.setParameter(KEY_EVENT_TYPE, EVENT_BLOCK_BREAK);
            msg.setPayload(frame.toJson());
            logger.trace("Sending message: {}", msg);
            sender.send(msg);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @EventHandler(ignoreCancelled = true)
    public void onBlockPlaced(BlockPlaceEvent event) {
        try {
            TelemetryFrame frame = new TelemetryFrame(Instant.now());
            addSrcLocation(frame, event.getPlayer().getLocation());
            addDstLocation(frame, event.getBlock().getLocation());
            addValue(frame, KEY_BLOCK_TYPE, event.getBlock().getType().toString());

            ESPMessage msg = new ESPMessage();
            msg.setParameter(KEY_SERVER, serverName);
            msg.setParameter(KEY_PLAYER_NAME, event.getPlayer().getName());
            msg.setParameter(KEY_WORLD, event.getBlock().getWorld().getName());
            msg.setParameter(KEY_EVENT_TYPE, EVENT_BLOCK_PLACE);
            msg.setPayload(frame.toJson());
            logger.trace("Sending message: {}", msg);
            sender.send(msg);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @EventHandler(ignoreCancelled = true)
    public void onPlayerMove(PlayerMoveEvent event) {
        try {
            TelemetryFrame frame = new TelemetryFrame(Instant.now());
            Location loc = event.getTo();
            addSrcLocation(frame, event.getFrom());
            addSrcAngle(frame, event.getFrom());
            addDstLocation(frame, event.getTo());
            addDstAngle(frame, event.getTo());

            ESPMessage msg = new ESPMessage();
            msg.setParameter(KEY_SERVER, serverName);
            msg.setParameter(KEY_PLAYER_NAME, event.getPlayer().getName());
            msg.setParameter(KEY_WORLD, loc.getWorld().getName());
            msg.setParameter(KEY_EVENT_TYPE, EVENT_PLAYER_MOVE);
            msg.setPayload(frame.toJson());
            sender.send(msg);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
