package lc.telemetry.faketsdb.service;

import lc.esp.sdk.*;
import lc.esp.sdk.telemetry.TelemetryFrame;
import lc.esp.sdk.telemetry.TelemetrySymbol;
import lc.mecha.fabric.HandlerStatus;
import lc.mecha.http.server.PrefixedHandler;
import lc.mecha.http.server.WebTransaction;
import lc.mecha.json.JSONArray;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import org.apache.commons.io.IOUtils;

import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.Set;

/**
 * /api/put handler to receive OpenTSDB messages and relay them into ESP.
 *
 * @author Alex Leigh
 * @since mk1 (HORIZON BRAVE)
 */
public class TSDBRelay extends PrefixedHandler {
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(TSDBRelay.class);
    public static final String KEY_TIMESTAMP = "timestamp";
    public static final String KEY_METRIC = "metric";
    public static final String KEY_TAGS = "tags";
    public static final String KEY_VALUE = "value";
    private final ESPProducer producer;

    public TSDBRelay(Set<String> pathPrefixes) throws Exception {
        super(pathPrefixes);
        ESPClient esp = new ESPClient();
        esp.start();
        ESPAddress telemDest = new ESPAddress("lc", "global",
                "env", "mon", ESPAddressClass.TOPIC,
                ESPMessageClass.TELEMETRY);
        ESPSession session = esp.createSession();
        producer = session.createProducer(telemDest);
    }

    @Override
    public HandlerStatus handlePrefixedWebRequest(WebTransaction request) throws Throwable {
        try {
            String msgStr = IOUtils.toString(request.httpServletRequest.getInputStream());
            JSONArray tsbMetrics = new JSONArray(msgStr);
            // System.out.println("Received opentsdb request: " + tsbMetrics);

            // For reach metric we are going to make a frame with a single symbol because our code is too
            // ignorant to sort by like tags
            for (int i = 0; i < tsbMetrics.length(); i++) {
                JSONObject tsbMetric = tsbMetrics.getJSONObject(i);
                TelemetryFrame frame = new TelemetryFrame(Instant.ofEpochSecond(tsbMetric.getLong(KEY_TIMESTAMP)));
                JSONObject tsdbTags = tsbMetric.getJSONObject(KEY_TAGS);
                for (String tagName : tsdbTags.keySet()) {
                    String tagValue = tsdbTags.getString(tagName);
                    frame.getTags().put(tagName, tagValue);
                }
                frame.getSymbols().add(new TelemetrySymbol(tsbMetric.getString(KEY_METRIC),
                        (Serializable) tsbMetric.get(KEY_VALUE),
                        null, false, true));

                logger.info("Built frame: {}", frame);

                // FIXME: Not threadsafe?
                ESPMessage telemMsg = frame.toMessage();
                producer.send(telemMsg);
            }

            JSONObject reply = new JSONObject();
            reply.put("success", tsbMetrics.length());
            reply.put("failed", 0);
            reply.put("errors", new JSONArray());

            request.httpServletResponse.getOutputStream().write(reply.toString().getBytes(StandardCharsets.UTF_8));

            return HandlerStatus.BREAK;
        } catch (Exception e) {
            e.printStackTrace();
            throw e;
        }
    }
}
