package lc.telemetry.historian.service;

import lc.esp.sdk.*;
import lc.esp.sdk.telemetry.TelemetryFrame;
import lc.esp.sdk.telemetry.TelemetrySymbol;
import lc.mecha.json.JSONArray;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import lc.mecha.util.BasicallyDangerous;
import lc.mecha.util.StringAccumulatorV2;
import lc.mecha.util.UniversalJob;
import lc.mecha.util.VelocityWatch;
import lc.zero.sdk.ZeroClient;
import lc.zero.sdk.ZeroServiceConfig;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.util.EntityUtils;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * The historian service receives telemetry messages from the ESP
 * message fabric and relays those messages to an OpenTSDB server.
 *
 * @author Alex Leigh
 * @since mk18 (GIPSY DANGER)
 */
public class HistorianService extends BasicallyDangerous {
    public static final String ZERO_SERVICE = "historian";
    public static final String TSDB_KEY_METRIC = "metric";
    public static final String TSDB_KEY_TIMESTAMP = "timestamp";
    public static final String TSDB_KEY_VALUE = "value";
    public static final String TSDB_KEY_TAGS = "tags";
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(HistorianService.class);
    private final CloseableHttpClient httpClient;
    private final ThreadPoolExecutor exec = (ThreadPoolExecutor) Executors.newCachedThreadPool();
    private final VelocityWatch vw = new VelocityWatch(logger);
    private String opentsdbUrl;

    public static void main(String[] args) {
        UniversalJob.banner(logger, "Historian mk1 (HORIZON BRAVE)");
        new HistorianService().run();
    }

    public HistorianService() {
        PoolingHttpClientConnectionManager connectionManager = new PoolingHttpClientConnectionManager();
        connectionManager.setMaxTotal(100);
        connectionManager.setDefaultMaxPerRoute(100);
        this.httpClient = HttpClients.custom().setConnectionManager(connectionManager).build();
    }

    @Override
    public void runDangerously() throws Exception {
        try (ESPClient esp = new ESPClient()) {
            esp.start();

            ZeroClient zero = esp.getZero();
            ZeroServiceConfig cfg = zero.getZai().readConfig(ZERO_SERVICE);
            if (cfg == null) {
                logger.error("No ZERO service configuration provided.");
                System.exit(UniversalJob.RET_BADENV);
            }
            logger.info("Found service configuration: {}", cfg);

            String addrUrl = cfg.getCfg().getString("address");
            ESPAddress telemDest = new ESPAddress(addrUrl);
            logger.info("Found address: {}", telemDest);

            opentsdbUrl = cfg.getCfg().getString("opentsdb");

            try (ESPSession session = esp.createSession()) {
                try (ESPConsumer consumer = session.createConsumer(telemDest)) {
                    //noinspection InfiniteLoopStatement
                    while (true) {
                        ESPMessage msg = consumer.receive();
                        publish(msg);
                    }
                }
            }
        }
    }

    /**
     * Publish the given telemetry message to OpenTSDB.
     */
    private void publish(ESPMessage msg) {
        exec.execute(new BasicallyDangerous() {
            @Override
            public void runDangerously() throws IOException {
                HttpPost httpPost = new HttpPost(opentsdbUrl);
                TelemetryFrame tf = new TelemetryFrame(msg);

                vw.event("frame");

                logger.trace("Built frame: {}", tf);

                // opentsdb only supports seconds precision
                long seconds = tf.getTime().getEpochSecond();

                JSONArray arr = new JSONArray();

                for (TelemetrySymbol symbol : tf.getSymbols()) {
                    JSONObject symJson = new JSONObject();

                    StringAccumulatorV2 sa = new StringAccumulatorV2(".");
                    sa.push(symbol.getName());
                    // TODO: We need to enforce formatting here
                    String uom = symbol.getUom();
                    if (uom != null) {
                        symJson.put(TSDB_KEY_METRIC, sa.asString() + "-" + uom);
                    } else {
                        symJson.put(TSDB_KEY_METRIC, sa.asString());
                    }

                    symJson.put(TSDB_KEY_TIMESTAMP, seconds);
                    symJson.put(TSDB_KEY_VALUE, symbol.getValue());
                    JSONObject tagsJson = new JSONObject();
                    symJson.put(TSDB_KEY_TAGS, tagsJson);

                    for (Map.Entry<String, Serializable> tag : tf.getTags().entrySet()) {
                        tagsJson.put(tag.getKey(), tag.getValue());
                    }

                    arr.put(symJson);
                }

                // OpenTSDB does not support chunked requests!
                BasicHttpEntity requestEntity = new BasicHttpEntity();
                byte[] r = arr.toString().getBytes(StandardCharsets.UTF_8);
                logger.trace("Built TSDB message: {}", arr.toString());

                requestEntity.setContent(new ByteArrayInputStream(r));
                requestEntity.setContentLength(r.length);
                requestEntity.setChunked(false);

                httpPost.setEntity(requestEntity);
                HttpResponse httpResponse = httpClient.execute(httpPost);

                // TSDB is very terse, it responds 204 on success
                /*
                if (httpResponse.getStatusLine().getStatusCode() != 204) {
                    System.out.println(httpResponse.getStatusLine());
                    Scanner sc = new Scanner(httpResponse.getEntity().getContent());
                    while (sc.hasNext()) {
                        System.out.println(sc.nextLine());
                    }
                }

                 */

                EntityUtils.consumeQuietly(httpResponse.getEntity());
            }
        });
        logger.trace("Received message: {}", msg);
    }
}
