package lc.zero.sdk;

import lc.mecha.http.client.HttpUtil;
import lc.mecha.json.JSONObject;
import lc.mecha.log.MechaLogger;
import lc.mecha.log.MechaLoggerFactory;
import lc.mecha.util.UniversalJob;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;

/**
 * Implementation of a ZERO client.
 * <p>
 * The client will firstly look for an environment variable <code>LC_ZAI_URL</code>. If specified it will use this
 * for the ZAI URL and not attempt network discovery. If this variable is unset, ZeroClient will listen for a
 * Zero Discovery Protocol (ZDP) packet containing a ZAI URL.
 *
 * @author Alex Leigh
 * @since mk1 (PUMA REAL)
 */
public class ZeroClient implements Runnable {
    public static final String ENV_LC_ZAI = "LC_ZERO_ZAI_URL";
    private static final MechaLogger logger = MechaLoggerFactory.getLogger(ZeroClient.class);
    public static int ZERO_PORT = 52000;
    private final HttpClient httpClient;
    private String zaiUrl = null;
    private ZeroActivationIndex zai;

    public ZeroClient() throws KeyStoreException, NoSuchAlgorithmException, KeyManagementException {
        this.httpClient = HttpUtil.makeOverlyTrustingClient(null);
    }

    @Override
    public void run() {
        String zai = System.getenv(ENV_LC_ZAI);
        if (StringUtils.isNotEmpty(zai)) {
            try {
                reconfigure(zai);
                return;
            } catch (Throwable e) {
                logger.error(ENV_LC_ZAI + " defined, but an error occurred: " + e);
                System.exit(UniversalJob.RET_BADENV);
            }
        } else {
            logger.info(ENV_LC_ZAI + " unset. Will perform network discovery.");
        }

        try {
            DatagramSocket serverSocket = new DatagramSocket(ZERO_PORT);
            serverSocket.setBroadcast(true);
            serverSocket.setReuseAddress(true);
            byte[] receiveData = new byte[1500];
            //noinspection InfiniteLoopStatement

            while (true) {
                DatagramPacket receivePacket = new DatagramPacket(receiveData, receiveData.length);
                serverSocket.receive(receivePacket);
                //  HexDump.hexDump(receivePacket.getData(), 0, receivePacket.getLength());
                // FIXME: If this throws we'll break the loop
                Beacon b = new Beacon(receivePacket.getData());
                processBeacon(b);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public ZeroActivationIndex getZai() {
        while (true) {
            if (zai != null) return zai;
            logger.info("Waiting for ZAI discovery...");
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    private void processBeacon(Beacon b) throws IOException {
        logger.debug("Received beacon: {}", b);
        if (this.zaiUrl == null) {
            reconfigure(b.getZaiUrl());
        } else if (this.zaiUrl.equalsIgnoreCase(b.getZaiUrl()) && b.isAuthoritative()) {
            reconfigure(b.getZaiUrl());
        }
    }

    /**
     * Initiate a reconfiguration to a new activation index URL.
     *
     * @param zaiUrl The activation index URL.
     */
    private void reconfigure(String zaiUrl) throws IOException {
        this.zaiUrl = zaiUrl;
        logger.info("Reconfiguration event. [ZAI: {}]", this.zaiUrl);
        HttpResponse response = httpClient.execute(new HttpGet(this.zaiUrl));
        HttpEntity entity = response.getEntity();
        JSONObject json = new JSONObject(entity.getContent());
        zai = new ZeroActivationIndex(httpClient, json);
        logger.info("Read ZAI: {}", zai);
    }
}